"""Metrics entities module"""

import os
import tempfile
from time import perf_counter
from typing import Literal, Protocol

# pylint: disable=R0801
if not os.environ.get("PROMETHEUS_MULTIPROC_DIR"):
    TEMPDIR = tempfile.TemporaryDirectory()  # pylint: disable=R1732
    MULTIPROC_DIR = TEMPDIR.name
    os.environ["PROMETHEUS_MULTIPROC_DIR"] = TEMPDIR.name
if not os.environ.get("PROMETHEUS_DISABLE_CREATED_SERIES"):
    os.environ["PROMETHEUS_DISABLE_CREATED_SERIES"] = "true"
# pylint: disable=C0413
from prometheus_client import CollectorRegistry, Counter, Histogram

from ..errors.errors import ErrorInfo

# Time series that contains count of observations that was less than or equal to its value for a given request time.
# Example:
# For buckets (0.1, 0.25, 0.5) and 2 requests by 0.2 sec and 3 requests by 0.4 sec observations will be:
#   +--------+-------+
#   | bucket | count |
#   +========+=======+
#   | 0.1    | 0     |
#   +--------+-------+
#   | 0.25   | 2     |
#   +--------+-------+
#   | 0.5    | 5     |
#   +--------+-------+
BUCKETS = (
    0.0001,
    0.00025,
    0.0005,
    0.001,
    0.0025,
    0.005,
    0.01,
    0.025,
    0.05,
    0.075,
    0.1,
    0.25,
    0.5,
    0.75,
    1.0,
    2.5,
    5.0,
    7.5,
    10.0,
)


class BaseMetricsContainer:
    """
    Base class for Metrics manager properties
    Attributes:
        requestLabels: Label names for the request metrics.
        errorLabels: Label names for the error metrics.
        errorCount: Counter for error metrics.
        requestCount: Counter for request metrics.
        extraValues: Label values for both metrics.
        multiprocEnabled: Whether or not to enable multiprocessing mode.
    """

    __slots__ = (
        "requestLabels",
        "errorLabels",
        "errorCount",
        "requestCount",
        "histogram",
        "extraValues",
        "multiprocEnabled",
    )

    def __init__(self, extraLabels: dict[str, str] | None = None):
        extraLabels = extraLabels or {}
        self.errorLabels = ["error_code", "path", *extraLabels.keys()]
        self.requestLabels = ["status_code", "path", *extraLabels.keys()]
        self.extraValues = [*extraLabels.values()]
        self.errorCount = None
        self.requestCount = None
        self.histogram = None
        self.multiprocEnabled = True

    def initialize(self, registry: CollectorRegistry):
        """
        Initialize the counters.
        Args:
            registry: Prometheus collector registry
        """
        extraArgs = {} if self.multiprocEnabled else {"registry": registry}

        self.errorCount = Counter("request_errors", "Counter of request errors", self.errorLabels, **extraArgs)
        self.requestCount = Counter("request_count", "Counter of requests", self.requestLabels, **extraArgs)
        self.histogram = Histogram(
            "requests", "Histogram of request time metrics", ["route", "status_code"], buckets=BUCKETS
        )

    def handleRequest(self, request, response):
        """Add point data to metrics"""
        path = f"{request.method}:{request.beautyRoute}"
        self.requestCount.labels(response.status, path, *self.extraValues).inc()
        self.histogram.labels(path, response.status).observe(perf_counter() - request.startRequestTimeCheckPoint)
        if isinstance(request.error, ErrorInfo):
            self.errorCount.labels(request.error.errorCode, path, *self.extraValues).inc()


class MetricsSettings(Protocol):
    """Metrics settings protocol"""

    # whether metrics are enabled
    enabled: Literal[0, 1]
    # metrics output format
    metricsFormat: Literal["prometheus"]
    # whether multiprocessing is enabled
    multiprocEnabled: Literal[0, 1]
