"""
Module implement license point class for monitoring
"""

from typing import Optional, Union

from .points import BaseMonitoringPoint


class BaseLicenseMonitoringPoint(BaseMonitoringPoint):
    """
    Base class for points which is associated with licensing.

    Attributes:
        licenseStatus (str): license status ("ok", "warning", "error")
        warnings: Optional[list[str]]: list of warning messages
        errors: Optional[list[str]]: list of error messages
    """

    #: monitoring series name
    series: str = "licensing"
    #: service name
    service: str

    __slots__ = ("licenseStatus", "warnings", "errors")

    @classmethod
    def initialize(cls, serviceName: str):
        """
        Initialize
        Args:
            serviceName: service name
        """
        cls.service = serviceName

    def __init__(
        self, eventTime: float, licenseStatus: str, warnings: Optional[list[str]], errors: Optional[list[str]]
    ):
        super().__init__(eventTime)
        self.licenseStatus = licenseStatus
        self.warnings = warnings
        self.errors = errors

    @property
    def tags(self) -> dict[str, Union[int, float, str]]:
        """
        Get tags

        Returns:
            dict with following keys: "service", "license_status"
        """
        return {"service": self.service, "license_status": self.licenseStatus}

    @property
    def fields(self) -> dict[str, Union[int, float, str]]:
        """
        Get fields

        Returns:
            dict with following keys: "warnings", "errors"
        """
        return {
            "warnings": " ".join(self.warnings) if self.warnings else None,
            "errors": " ".join(self.errors) if self.errors else None,
        }
