"""
Module realizes service handler
"""

from abc import ABC, abstractmethod
from contextlib import asynccontextmanager
from copy import copy
from typing import Any, AsyncGenerator, Callable, Generic, Mapping, Optional, Set, Tuple, Type, TypeVar, Union

from sanic.response import HTTPResponse
from sanic.views import HTTPMethodView
from vlutils.helpers import isUUID
from vlutils.structures.pydantic import BaseModel

from ..errors.errors import Error, ErrorInfo
from ..errors.exception import VLException
from ..pydantic.loaders import loadDataFromJson, loadDataFromQuery
from ..web import response
from .query_getters import validStringGetter
from .request import CONFIG_CLASS, VLRequest
from .utils import validateJson

# (List[str]): all support http methods
ALLOW_METHODS = ["GET", "PUT", "OPTIONS", "HEAD", "POST", "DELETE", "PATCH"]
# (List[str]): valid content type for json requests
VALID_REQUEST_JSON_CONTENT_TYPES = ["application/json", "application/octet-stream"]

REQUEST_CLASS = TypeVar("REQUEST_CLASS", bound=VLRequest)


class BaseHandler(Generic[REQUEST_CLASS, CONFIG_CLASS], HTTPMethodView):
    """
    Base handler
        logger(Logger): Logger object
        requestId(str): RequestId consists of two parts. first - timestamp of server time or utc, second - uuid
        respHeaders(dict): dict for collecting Response headers
        request (VLRequest): input request
    """

    def __init__(self, request: REQUEST_CLASS):
        super().__init__()
        self.request = request
        self.logger = request.logger
        self.requestId = request.requestId
        self.respHeaders = {}

    def setMethodNotAllowed(self) -> HTTPResponse:
        """
        Set error "method not allowed" and status code 405
        """
        return self.error(405, error=Error.MethodNotAllowed)

    def dispatchRequest(self, request: VLRequest, *args, **kwargs):
        """
        Overload dispatch_request, remove `request` fro handler method arguments
        Args:
            request: request
            *args: args
            **kwargs: kwargs

        Returns:
            handler method call
        """
        handler = getattr(self, request.method.lower(), None)
        if handler is None:
            raise VLException(Error.MethodNotAllowed, statusCode=405, isCriticalError=False)
        return handler(*args, **kwargs)

    @classmethod
    def as_view(cls, *class_args, **class_kwargs):  # pylint: disable-msg=C0103
        """Return view function for use with the routing system, that
        dispatches request to appropriate handler method.
        """

        def view(*args, **kwargs):
            self = view.view_class(args[0], *class_args, **class_kwargs)
            return self.dispatchRequest(*args, **kwargs)

        if cls.decorators:
            view.__module__ = cls.__module__
            for decorator in cls.decorators:
                view = decorator(view)

        view.view_class = cls
        view.__doc__ = cls.__doc__
        view.__module__ = cls.__module__
        view.__name__ = cls.__name__
        return view

    def error(self, statusCode: int, error: ErrorInfo) -> HTTPResponse:
        """
        Function, that creates error response. Calls installation of request result.

        Args:
            statusCode: response status code
            error:  error
        """

        reply = response.json(status=statusCode, body=error.asDict(), headers=self.respHeaders)
        self.request.error = error
        return reply

    def success(
        self,
        statusCode: int = 200,
        body: Optional[Union[bytes, str]] = None,
        outputJson: Optional[Union[dict, list]] = None,
        contentType: Optional[str] = None,
        extraHeaders: Optional[Mapping[str, str]] = None,
    ) -> HTTPResponse:
        """
        Finish  success request. Generate correct reply with request id header, correct  Content type header

        Args:
            contentType: body content type
            statusCode: response status code, range(200, 300), default 200
            body: pure body
            outputJson: json as object
            extraHeaders: extra headers that will be added to the response
                (default headers in case of overlapping will be replaced);
        Returns:
            aiohttp.web.Response object
        Raises:
            ValueError: if response has body but its content-type was not determined in "Content-Type" header

        """

        headers = self.respHeaders
        if contentType is not None:
            headers["Content-Type"] = contentType
        if extraHeaders is not None:
            headers.update(extraHeaders)

        if outputJson is not None:
            reply = response.json(body=outputJson, headers=headers, status=statusCode)
        elif body is not None:
            if "Content-Type" not in headers:
                raise ValueError("Response has body but does not have proper Content-Type header")
            reply = response.VLHTTPResponse(body=body, headers=headers, content_type=contentType)
        else:
            if "Content-Type" not in headers:
                headers["Content-Type"] = "application/octet-stream"
            reply = response.VLHTTPResponse(headers=headers, content_type=contentType)
        reply.status = statusCode
        return reply

    async def stream(self, producer: AsyncGenerator, contentType: str, headers: dict | None = None, status: int = 200):
        """
        Interface for streaming data with async_generator.

        Args:
            producer: async_generator that produce bytes
            contentType: content-type of data
            headers: response headers
            status: response status code
        """
        resp = await self.request.respond(content_type=contentType, headers=headers, status=status)
        async for chunk in producer:
            await resp.send(chunk)
        await resp.eof()

    def validateJson(
        self, data: Union[int, dict, list], schema: Union[dict, Callable], useJsonSchema: bool = False
    ) -> None:
        """
        Validate json or msgpack.

        Args:
            data: json for validation
            schema: json schema if using jsonschema, compiled schema if using fastjsonschema
                    fastjsonschema doesn't full support dependencies in schemas at the moment
            useJsonSchema: flag to use jsonschema instead of fastjsonschema (don't support dependencies at the moment)
        Raises:
            VLException(Error.BadInputJson, 400): if failed  to validate schema
            VLException(Error.BadInputMsgpack, 400): if failed  to validate schema
        """
        return validateJson(self.request.content_type, data=data, schema=schema, useJsonSchema=useJsonSchema)

    def addDataForMonitoring(self, tags: Optional[dict] = None, fields: Optional[dict] = None) -> None:
        """
        Add data to request monitoring

        Args:
            tags: tags (indexed  data)
            fields: fields (non indexed  data)
        """
        if tags is not None:
            self.request.dataForMonitoring.tags.update(tags)
        if fields is not None:
            self.request.dataForMonitoring.fields.update(fields)

    def getQueryParam(
        self, name: str, validator: Callable = validStringGetter, require: bool = False, default: Any = None
    ) -> Any:
        """
        Getting query param

        Args:
            name: name of param
            validator: param validator and converter from str to correct format
            require: whether param is required
            default: default value of param
        Returns:
            value
        Raises:
            VLException(Error.QueryParameterNotFound, 400): if require param not found
            VLException(Error.BadQueryParams, 400): if format of param is incorrect
        """
        valueStr = self.request.args.get(name)
        if valueStr is None:
            if not require:
                if default:
                    return copy(default)
                return default
            error = Error.QueryParameterNotFound.format(name)
            raise VLException(error, 400, False)

        try:
            return validator(valueStr)
        except ValueError as exc:
            error = Error.BadQueryParams.format(name)
            raise VLException(error, 400, False) from exc

    def getPagination(
        self, defaultPage: int = 1, defaultPageSize: int = 10, maxPageSize: int = 1000, minPageSize: int = 1
    ) -> Tuple[int, int]:
        """
        Getting pagination

        Args:
            maxPageSize: max elements by page
            minPageSize: min elements by page
            defaultPage: default page
            defaultPageSize: default page size
            pair page and page_size
        Returns:
            pair page and page size
        Raises:
            VLException(Error.BadQueryParams.format("page, should be less 2^32"), 400, isCriticalError=False) if page
                                                                                                    is greater than 2^32
        """
        page = self.getQueryParam("page", lambda params: max(int(params), 1), default=defaultPage)
        if page > 2**32:
            raise VLException(Error.BadQueryParams.format("page, should be less than 2^32"), 400, isCriticalError=False)
        pageSize = self.getQueryParam(
            "page_size", lambda params: max(min(int(params), maxPageSize), minPageSize), default=defaultPageSize
        )
        return page, pageSize

    def getAllowedMethods(self) -> Set[str]:
        """
        Get Allowed methods for resource

        Returns:
            set object with supported methods
        """
        handlerMethods = getattr(self, "allowedMethods", None)
        if handlerMethods:
            allowedMethods = set(handlerMethods)
        else:
            allowedMethods = set()

        if not allowedMethods:
            for method in ALLOW_METHODS:
                if hasattr(self, method.lower()):
                    if getattr(self, method.lower()).__qualname__.startswith(self.__class__.__name__):
                        allowedMethods.add(method)
        allowedMethods.add("OPTIONS")
        return allowedMethods

    def getAccountIdFromHeader(self, required: bool = True) -> Optional[str]:
        """
        Get required account id from header Luna-Account-Id
        Args:
            required: whether header must be specified
        Returns:
            account id if exists
        """
        accountId = self.request.headers.get("Luna-Account-Id")
        if required and accountId is None:
            raise VLException(Error.LunaAccountIdHeaderNotFound, 400, isCriticalError=False)
        if accountId is not None and not isUUID(accountId):
            raise VLException(Error.BadAccountId, 400, isCriticalError=False)
        return accountId

    @property
    def app(self):
        """
        Get sanic application.
        Returns:
            app
        """
        return self.request.app

    @property
    def config(self) -> CONFIG_CLASS:
        """
        Get app config
        Returns:
            app config
        """
        return self.request.appCtx.serviceConfig

    @staticmethod
    def loadDataFromJson(data: dict, model: Type[BaseModel]) -> Any:
        """
        Load data from json with pydantic
        Args:
            data: input data
            model: pydantic model

        Returns:
            initialized object
        """
        return loadDataFromJson(data, model)

    def loadDataFromQuery(self, model: Type[BaseModel]) -> Any:
        """
        Load data from query with pydantic
        Args:
            model: pydantic model

        Returns:
            initialized object
        """
        return loadDataFromQuery({k: self.request.args.get(k) for k in self.request.args.keys()}, model)


class WSBaseHandler(BaseHandler, ABC):
    """
    Base class for websocket handlers
    """

    @asynccontextmanager
    @abstractmethod
    async def wsSession(self, request: VLRequest):
        """
        Context manager for  ws connection before handshake.
        Validation and preparation for ws request must be here.
        """
